# Base class for tasks
from collections import OrderedDict as odict
import threading
from datetime import datetime
import os
import logging
import tables
# from autopilot.core.networking import Net_Node
from autopilot.hardware import BCM_TO_BOARD
from autopilot import prefs
from autopilot.core.loggers import init_logger

if prefs.get( "AUDIOSERVER"):
    if prefs.get('AUDIOSERVER') == 'pyo':
        pass
    elif prefs.get('AUDIOSERVER') == 'jack':
        pass


class Task(object):
    """
    Generic Task metaclass

    Attributes:
        PARAMS (:class:`collections.OrderedDict`): Params to define task, like::

            PARAMS = odict()
            PARAMS['reward']         = {'tag':'Reward Duration (ms)',
                                        'type':'int'}
            PARAMS['req_reward']     = {'tag':'Request Rewards',
                                        'type':'bool'}

        HARDWARE (dict): dict for necessary hardware, like::

            HARDWARE = {
                'POKES':{
                    'L': hardware.Beambreak, ...
                },
                'PORTS':{
                    'L': hardware.Solenoid, ...
                }
            }

        PLOT (dict): Dict of plotting parameters, like::

                PLOT = {
                    'data': {
                        'target'   : 'point',
                        'response' : 'segment',
                        'correct'  : 'rollmean'
                    },
                    'chance_bar'  : True, # Draw a red bar at 50%
                    'roll_window' : 50 # number of trials to roll window over
                }

        Trial_Data (:class:`tables.IsDescription`): Data table description, like::

            class TrialData(tables.IsDescription):
                trial_num = tables.Int32Col()
                target = tables.StringCol(1)
                response = tables.StringCol(1)
                correct = tables.Int32Col()
                correction = tables.Int32Col()
                RQ_timestamp = tables.StringCol(26)
                DC_timestamp = tables.StringCol(26)
                bailed = tables.Int32Col()

        STAGE_NAMES (list): List of stage method names
        stage_block (:class:`threading.Event`): Signal when task stages complete.
        punish_stim (bool): Do a punishment stimulus
        stages (iterator): Some generator or iterator that continuously returns the next stage method of a trial
        triggers (dict): Some mapping of some pin to callback methods
        pins (dict): Dict to store references to hardware
        pin_id (dict): Reverse dictionary, pin numbers back to pin letters.
        punish_block (:class:`threading.Event`): Event to mark when punishment is occuring
        logger (:class:`logging.Logger`): gets the 'main' logger for now.
    """
    # dictionary of Params needed to define task,
    # these should correspond to argument names for the task
    PARAMS = odict()

    # Task Definition
    HARDWARE = {} # Hardware needed to run the task
    STAGE_NAMES = [] # list of names of stage methods
    PLOT = {} # dictionary of plotting params
    class TrialData(tables.IsDescription):
        trial_num = tables.Int32Col()
        session = tables.Int32Col()



    def __init__(self, *args, **kwargs):

        # Task management
        self.stage_block = None  # a threading.Event used by the pilot to manage stage transitions
        self.punish_stim = False
        #self.running = None  # Event used to exit the run thread
        self.stages = None  # Some generator that continuously returns the next stage of the trial
        self.triggers = {}
        self.stim_manager = None

        self.trial_counter = None  # will be init'd by the subtask because will use the current trial

        # Hardware
        self.hardware = {}  # dict to store references to hardware
        self.pin_id = {}  # pin numbers back to pin lettering

        self.punish_block = threading.Event()
        self.punish_block.set()
        #self.running = threading.Event()

        # try to get logger
        self.logger = init_logger(self)

    def init_hardware(self):
        """
        Use the HARDWARE dict that specifies what we need to run the task
        alongside the HARDWARE subdict in :mod:`prefs` to tell us how
        they're plugged in to the pi

        Instantiate the hardware, assign it :meth:`.Task.handle_trigger`
        as a callback if it is a trigger.
        """
        # We use the HARDWARE dict that specifies what we need to run the task
        # alongside the HARDWARE subdict in the prefs structure to tell us how they're plugged in to the pi
        self.hardware = {}
        self.pin_id = {} # Reverse dict to identify pokes
        pin_numbers = prefs.get('HARDWARE')

        # We first iterate through the types of hardware we need
        for type, values in self.HARDWARE.items():
            self.hardware[type] = {}
            # then iterate through each pin and handler of this type
            for pin, handler in values.items():
                try:
                    hw_args = pin_numbers[type][pin]
                    if isinstance(hw_args, dict):
                        if 'name' not in hw_args.keys():
                            hw_args['name'] = "{}_{}".format(type, pin)
                        hw = handler(**hw_args)
                    else:
                        hw_name = "{}_{}".format(type, pin)
                        hw = handler(hw_args, name=hw_name)

                    # if a pin is a trigger pin (event-based input), give it the trigger handler
                    if hw.is_trigger:
                        hw.assign_cb(self.handle_trigger)

                    # add to forward and backwards pin dicts
                    self.hardware[type][pin] = hw
                    if isinstance(hw_args, int) or isinstance(hw_args, str):
                        self.pin_id[hw_args] = pin
                    elif isinstance(hw_args, list):
                        for p in hw_args:
                            self.pin_id[p] = pin
                    elif isinstance(hw_args, dict):
                        if 'pin' in hw_args.keys():
                            self.pin_id[hw_args['pin']] = pin 

                except:
                    self.logger.exception("Pin could not be instantiated - Type: {}, Pin: {}".format(type, pin))

    def set_reward(self, vol=None, duration=None, port=None):
        """
        Set the reward value for each of the 'PORTS'.

        Args:
            vol(float, int): Volume of reward in uL
            duration (float): Duration to open port in ms
            port (None, Port_ID): If `None`, set everything in 'PORTS', otherwise
                only set `port`
        """
        if not vol and not duration:
            raise Exception("Need to have duration or volume!!")
        if vol and duration:
            self.logger.warning('given both volume and duration, using volume.')

        if not port:
            for k, port in self.hardware['PORTS'].items():
                if vol:
                    try:
                        port.dur_from_vol(vol)
                    except AttributeError:
                        self.logger.warning('No calibration found, using duration = 20ms instead')
                        port.duration = 0.02
                else:
                    port.duration = float(duration)/1000.
        else:
            try:
                if vol:
                    try:
                        self.hardware['PORTS'][port].dur_from_vol(vol)
                    except AttributeError:
                        self.logger.warning('No calibration found, using duration = 20ms instead')
                        port.duration = 0.02

                else:
                    self.hardware['PORTS'][port].duration = float(duration) / 1000.
            except KeyError:
                raise Exception('No port found named {}'.format(port))

    def handle_trigger(self, pin, level=None, tick=None):
        """
        All GPIO triggers call this function with the pin number, level (high, low),
        and ticks since booting pigpio.

        Calls any trigger assigned to the pin in `self.triggers` ,
        unless during punishment (returns).

        Args:
            pin (int): BCM Pin number
            level (bool): True, False high/low
            tick (int): ticks since booting pigpio
        """
        # All triggers call this function with the pin number, level (high, low), and ticks since booting pigpio

        # We get fed hardware as BCM numbers, convert to board number and then back to letters
        if isinstance(pin, int):
            pin = BCM_TO_BOARD[pin]
            pin = self.pin_id[pin]

        if pin not in self.triggers.keys():
            self.logger.debug(f"No trigger found for {pin}")
            # No trigger assigned, get out without waiting
            return

        if pin == 'TIMEUP':
            # TODO: Handle timers, reset trial
            # TODO: Handle bailing, for example by replacing the cycle with a single function that returns the 'bail' flag
            return

        # if we're being punished, don't recognize the trigger
        if not self.punish_block.is_set():
            return

        # Call the trigger
        try:
            self.triggers[pin]()
        except TypeError:
            # Multiple triggers, call them all
            for trig in self.triggers[pin]:
                trig()
        except KeyError:
            # If we don't have a trigger, that's fine, eg. L and R before requesting
            return

        # clear triggers
        self.triggers = {}

        # Set the stage block so the pilot calls the next stage
        self.stage_block.set()

    def set_leds(self, color_dict=None):
        """
        Set the color of all LEDs at once.

        Args:
            color_dict (dict): If None, turn LEDs off, otherwise like:

                {'pin': [R,G,B],
                'pin2: [R,G,B]}


        """
        # We are passed a dict of ['pin']:[R, G, B] to set multiple colors
        # All others are turned off
        if not color_dict:
            color_dict = {}
        for k, v in self.hardware['LEDS'].items():
            if k in color_dict.keys():
                v.set(color_dict[k])
            else:
                v.set(0)

    def flash_leds(self):
        """
        flash lights for punish_dir
        """
        for k, v in self.hardware['LEDS'].items():
            if v.__module__ == "autopilot.hardware.gpio" and type(v).__name__ == "LED_RGB":
                v.flash(self.punish_dur)

    def end(self):
        """
        Release all hardware objects
        """
        for k, v in self.hardware.items():
            for pin, obj in v.items():
                obj.release()

        if hasattr(self, 'stim_manager'):
            if self.stim_manager is not None:
                self.stim_manager.end()












from autopilot import prefs
# if prefs.get('AGENT') in ("TERMINAL", "DOCS"):
HAVE_PYSIDE = False
try:
    from PySide2 import QtCore
    HAVE_PYSIDE = True
except ImportError:
    pass

from collections import OrderedDict as odict
import json
import pandas as pd
from scipy.stats import linregress
# from subprocess import call
from threading import Thread
import os
import numpy as np

class Param(object):
    """
    In the future, we will implement a coherent Parameter management system

    Warning:
        Not Implemented.
    """
    # Class to hold and verify task and gui parameters
    tag = None # human-readable description of parameter
    type = None # string that describes the type of input or param

    # possible types
    types = ['int', 'bool', 'list']

    def __init__(self, **kwargs):
        """
        Args:
            **kwargs:
        """
        for k, v in kwargs.items():
            setattr(self, k, v)

    # enable dictionary-like behavior
    def __getitem__(self, key):
        """
        Args:
            key:
        """
        return self.__dict__[key]

    def __setitem__(self, key, value):
        """
        Args:
            key:
            value:
        """
        self.__dict__[key] = value

    def __delitem__(self, key):
        """
        Args:
            key:
        """
        del self.__dict__[key]

    def __contains__(self, key):
        """
        Args:
            key:
        """
        return key in self.__dict__

    def __len__(self):
        return len(self.__dict__)

    # def validate(self):
    #     if all([self.id, self.to, self.sender, self.key]):
    #         return True
    #     else:
    #         return False


_INVOKER = None

if HAVE_PYSIDE:
    class InvokeEvent(QtCore.QEvent):
        """
        Sends signals to the main QT thread from spawned message threads

        See `stackoverflow <https://stackoverflow.com/a/12127115>`_
        """

        EVENT_TYPE = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

        def __init__(self, fn, *args, **kwargs):
            # type: (function, object, object) -> None
            """
            Accepts a function, its args and kwargs and wraps them as a
            :class:`QtCore.QEvent`

            """
            QtCore.QEvent.__init__(self, InvokeEvent.EVENT_TYPE)
            self.fn = fn
            self.args = args
            self.kwargs = kwargs


    class Invoker(QtCore.QObject):
        """
        Wrapper that calls an evoked event made by :class:`.InvokeEvent`
        """
        def event(self, event):
            """
            Args:
                event:
            """
            event.fn(*event.args, **event.kwargs)
            return True

    def get_invoker():
        if globals()['_INVOKER'] is None:
            globals()['_INVOKER'] = Invoker()
        return globals()['_INVOKER']


class ReturnThread(Thread):
    """
    Thread whose .join() method returns the value from the function
    thx to https://stackoverflow.com/a/6894023
    """
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs={}, Verbose=None):
        Thread.__init__(self, group, target, name, args, kwargs, Verbose)
        self._return = None
    def run(self):
        if self._Thread__target is not None:
            self._return = self._Thread__target(*self._Thread__args,
                                                **self._Thread__kwargs)
    def join(self, timeout=None):
        Thread.join(self, timeout)

        return self._return

def list_subjects(pilot_db=None):
    """
    Given a dictionary of a pilot_db, return the subjects that are in it.

    Args:
        pilot_db (dict): a pilot_db. if None tried to load pilot_db with :method:`.load_pilotdb`

    Returns:
        subjects (list): a list of currently active subjects

    """

    if pilot_db is None:
        pilot_db = load_pilotdb()

    subjects = []
    for pilot, values in pilot_db.items():
        if 'subjects' in values.keys():
            subjects.extend(values['subjects'])

    return subjects

def load_pilotdb(file_name=None, reverse=False):
    """
    Try to load the file_db

    Args:
        reverse (bool): Return inverted pilot db mapping subjects: pilots (default False)
        file_name (str): Path of ``pilot_db.json``, if None, use ``prefs.get('PILOT_DB')``

    Returns:
        :class:`collections.OrderedDict` : pilot_db.json or reversed pilot_db
    """

    if file_name is None:
        file_name = prefs.get('PILOT_DB')

    with open(file_name) as pilot_file:
        pilot_db = json.load(pilot_file, object_pairs_hook=odict)

    if reverse:
        # simplify pilot db
        pilot_db = odict({k: v['subjects'] for k, v in pilot_db.items()})
        pilot_dict = odict()
        for pilot, subjectlist in pilot_db.items():
            for ms in subjectlist:
                pilot_dict[ms] = pilot
        pilot_db = pilot_dict

    return pilot_db

def coerce_discrete(df, col, mapping={'L':0, 'R':1}):
    """
    Coerce a discrete/string column of a pandas dataframe into numeric values

    Default is to map 'L' to 0 and 'R' to 1 as in the case of Left/Right 2AFC tasks

    Args:
        df (:class:`pandas.DataFrame`) : dataframe with the column to transform
        col (str):  name of column
        mapping (dict): mapping of strings to numbers

    Returns:
        df (:class:`pandas.DataFrame`) : transformed dataframe

    """

    for key, val in mapping.items():
        df.loc[df[col]==key,col] = val

    # if blanks, warn and remove
    if '' in df[col].unique():
        n_blanks = sum(df[col]=='')
        Warning('{} blank rows detected, removing.'.format(n_blanks))
        df.drop(df.index[df[col]==''], axis=0, inplace=True)

    df = df.astype({col:float})
    return df


def find_recursive(key, dictionary):
    """
    Find all instances of a key in a dictionary, recursively.

    Args:
        key:
        dictionary:

    Returns:
        list
    """
    for k, v in dictionary.iteritems():
        if k == key:
            yield v
        elif isinstance(v, dict):
            for result in find_recursive(key, v):
                yield result
        elif isinstance(v, list):
            for d in v:
                for result in find_recursive(key, d):
                    yield result



#
# def update_pis(github=True, apt=False, pilot_select = None, prefs_fn = None):
#     """
#     Args:
#         github:
#         apt:
#         pilot_select:
#         prefs_fn:
#     """
#     # update github, or apt?
#     # should limit pilots or use all?
#     # load prefs from default location or use different?
#     if prefs_fn is None:
#         prefs = get_prefs()
#     else:
#         prefs = get_prefs(prefs_fn)
#
#     # get ips from pilot db
#     with open(prefs['PILOT_DB'], 'r') as pilot_db:
#         pilots = json.load(pilot_db)
#
#     # if we were passed a list of pilots to subset then do it
#     if pilot_select is not None:
#         pilots = {k: v for k, v in pilots.items() if k in pilot_select }
#
#     if github is True:
#         ips = ['pi@'+v['ip'] for k,v in pilots.items()]
#         ip_string = " ".join(ips)
#         call('parallel-ssh', '-H', ip_string, 'git --git-dir=/home/pi/git/autopilot/.git pull')










import subprocess
import os
import sys
from autopilot import prefs
import atexit
from time import sleep
import threading
import shutil
import signal

PIGPIO = False
PIGPIO_DAEMON = None
PIGPIO_LOCK = threading.Lock()
try:
    if shutil.which('pigpiod') is not None:
        PIGPIO = True

except ImportError:
    pass

JACKD = False
# JACKD_MODULE = None # Whether jackd is a module in autopilot.external (True) or use system jackd (False)
JACKD_PROCESS = None
try:
    import jack
    JACKD = True
    # from autopilot.external import jack as autopilot_jack
    #
    # # set env variables
    # jackd_path = os.path.join(autopilot_jack.__path__._path[0])
    #
    # # specify location of libraries when starting jackd
    #
    #
    # if 'LD_LIBRARY_PATH' in os.environ.keys():
    #
    #     os.environ['LD_LIBRARY_PATH'] = ":".join([os.path.join(jackd_path, 'lib'),
    #                                               os.environ.get('LD_LIBRARY_PATH',"")])
    # else:
    #     os.environ['LD_LIBRARY_PATH'] = os.path.join(jackd_path, 'lib')git pu
    # # lib_string = "LD_LIBRARY_PATH=" + os.path.join(jackd_path, 'lib')
    #
    # # specify location of drivers when starting jackd
    # os.environ['JACK_DRIVER_DIR'] = os.path.join(jackd_path, 'lib', 'jack')
    # # driver_string = "JACK_DRIVER_DIR=" + os.path.join(jackd_path, 'lib', 'jack')
    #
    # JACKD = True
    # JACKD_MODULE = True

except (ImportError, OSError):
    pass
    # # try to import jack client, it will look for system jack by default
    # try:
    #     import jack
    #     JACKD = True
    #     JACKD_MODULE = False
    # except OSError:
    #     # no need to do anything, just setting module variables that test what the system is configured to do
    #     pass


def start_pigpiod():
    if not PIGPIO:
        raise ImportError('the pigpiod daemon was not found! use autopilot.setup.')

    with globals()['PIGPIO_LOCK']:
        if globals()['PIGPIO_DAEMON'] is not None:
            return globals()['PIGPIO_DAEMON']

        launch_pigpiod = shutil.which('pigpiod')
        if launch_pigpiod is None:
            raise RuntimeError('the pigpiod binary was not found!')

        if prefs.get( 'PIGPIOARGS'):
            launch_pigpiod += ' ' + prefs.get('PIGPIOARGS')

        if prefs.get( 'PIGPIOMASK'):
            # if it's been converted to an integer, convert back to a string and zfill any leading zeros that were lost
            if isinstance(prefs.get('PIGPIOMASK'), int):
                prefs.set('PIGPIOMASK', str(prefs.get('PIGPIOMASK')).zfill(28))
            launch_pigpiod += ' -x ' + prefs.get('PIGPIOMASK')

        proc = subprocess.Popen('sudo ' + launch_pigpiod, shell=True)
        globals()['PIGPIO_DAEMON'] = proc

        # kill process when session ends
        def kill_proc(*args):
            proc.kill()
            sys.exit(1)
        atexit.register(kill_proc)
        signal.signal(signal.SIGTERM, kill_proc)

        # sleep to let it boot up
        sleep(1)

        return proc

def start_jackd():
    if not JACKD:
        raise ImportError('jackd was not found in autopilot.external or as a system install')

    # get specific launch string from prefs
    if prefs.get("JACKDSTRING"):
        jackd_string = prefs.get('JACKDSTRING').lstrip('jackd')

    else:
        jackd_string = ""

    # replace string fs with number
    if prefs.get('FS'):
        jackd_string = jackd_string.replace('-rfs', f"-r{prefs.get('FS')}")

    # construct rest of launch string!
    # if JACKD_MODULE:
    #     jackd_path = os.path.join(autopilot_jack.__path__._path[0])
    #
    #     # now set as env variables...
    #     # specify location of libraries when starting jackd
    #     # lib_string = "LD_LIBRARY_PATH=" + os.path.join(jackd_path, 'lib')
    #     #
    #     # specify location of drivers when starting jackd
    #     # os.environ['JACK_DRIVER_DIR'] = os.path.join(jackd_path, 'lib', 'jack')
    #     # driver_string = "JACK_DRIVER_DIR=" + os.path.join(jackd_path, 'lib', 'jack')
    #
    #     jackd_bin = os.path.join(jackd_path, 'bin', 'jackd')
    #
    #     # combine all the pieces
    #     # launch_jackd = " ".join([lib_string, driver_string, jackd_bin, jackd_string])
    #
    # else:
    jackd_bin = shutil.which('jackd')

    #jackd_bin = 'jackd'

        # launch_jackd = " ".join([jackd_bin, jackd_string])

    launch_jackd = " ".join([jackd_bin, jackd_string])

    proc = subprocess.Popen(launch_jackd, shell=True)
    globals()['JACKD_PROCESS'] = proc

    # kill process when session ends
    def kill_proc(*args):
        proc.kill()
        sys.exit(1)
    atexit.register(kill_proc)
    signal.signal(signal.SIGTERM, kill_proc)

    # sleep to let it boot
    sleep(2)

    return proc




"""
Classes to plot data in the GUI.

.. todo::

    Add all possible plot objects and options in list.

Note:
    Plot objects need to be added to :data:`~.plots.PLOT_LIST` in order to be reachable.


"""

# Classes for plots
import sys
import logging
import os
import numpy as np
import PySide2 # have to import to tell pyqtgraph to use it
import pandas as pd
from PySide2 import QtGui
from PySide2 import QtCore
from PySide2 import QtOpenGL
from PySide2 import QtWidgets
import pyqtgraph as pg
from time import time, sleep
from itertools import count
from functools import wraps
from threading import Event, Thread
import multiprocessing as mp
import pdb
from queue import Queue, Empty, Full
#import cv2
pg.setConfigOptions(antialias=True)
# from pyqtgraph.widgets.RawImageWidget import RawImageWidget, RawImageGLWidget

from autopilot import tasks, prefs
from autopilot.core import styles
from autopilot.core.utils import get_invoker
from .utils import InvokeEvent, Invoker
from autopilot.core.networking import Net_Node
from autopilot.core.loggers import init_logger


############
# Plot list at the bottom!
###########

def gui_event(fn):
    """
    Wrapper/decorator around an event that posts GUI events back to the main
    thread that our window is running in.

    Args:
        fn (callable): a function that does something to the GUI
    """
    @wraps(fn)
    def wrapper_gui_event(*args, **kwargs):
        # type: (object, object) -> None
        """

        Args:
            *args (): 
            **kwargs (): 
        """
        QtCore.QCoreApplication.postEvent(get_invoker(), InvokeEvent(fn, *args, **kwargs))
    return wrapper_gui_event


class Plot_Widget(QtWidgets.QWidget):
    """
    Main plot widget that holds plots for all pilots

    Essentially just a container to give plots a layout and handle any
    logic that should apply to all plots.

    Attributes:
        logger (`logging.Logger`): The 'main' logger
        plots (dict): mapping from pilot name to :class:`.Plot`
    """
    # Widget that frames multiple plots
    def __init__(self):
        # type: () -> None
        QtWidgets.QWidget.__init__(self)

        self.logger = init_logger(self)


        # We should get passed a list of pilots to keep ourselves in order after initing
        self.pilots = None

        # Dict to store handles to plot windows by pilot
        self.plots = {}

        # Main Layout
        self.layout = QtWidgets.QVBoxLayout(self)
        self.layout.setContentsMargins(0,0,0,0)
        self.layout.setSpacing(0)

        # Plot Selection Buttons
        # TODO: Each plot bar should have an option panel, because different tasks have different plots
        #self.plot_select = self.create_plot_buttons()

        # Create empty plot container
        self.plot_layout = QtWidgets.QVBoxLayout()

        # Assemble buttons and plots
        #self.layout.addWidget(self.plot_select)
        self.layout.addLayout(self.plot_layout)

        self.setLayout(self.layout)

        self.setContentsMargins(0, 0, 0, 0)

    def init_plots(self, pilot_list):
        """
        For each pilot, instantiate a :class:`.Plot` and add to layout.

        Args:
            pilot_list (list): the keys from :attr:`.Terminal.pilots`
        """
        self.pilots = pilot_list

        # Make a plot for each pilot.
        for p in self.pilots:
            plot = Plot(pilot=p, parent=self)
            self.plot_layout.addWidget(plot)
            self.plot_layout.addWidget(HLine())
            self.plots[p] = plot


class Plot(QtWidgets.QWidget):
    """
    Widget that hosts a :class:`pyqtgraph.PlotWidget` and manages
    graphical objects for one pilot depending on the task.

    **listens**

    +-------------+------------------------+-------------------------+
    | Key         | Method                 | Description             |
    +=============+========================+=========================+
    | **'START'** | :meth:`~.Plot.l_start` | starting a new task     |
    +-------------+------------------------+-------------------------+
    | **'DATA'**  | :meth:`~.Plot.l_data`  | getting a new datapoint |
    +-------------+------------------------+-------------------------+
    | **'STOP'**  | :meth:`~.Plot.l_stop`  | stop the task           |
    +-------------+------------------------+-------------------------+
    | **'PARAM'** | :meth:`~.Plot.l_param` | change some parameter   |
    +-------------+------------------------+-------------------------+

    **Plot Parameters**

    The plot is built from the ``PLOT={data:plot_element}`` mappings described in the :class:`~autopilot.tasks.task.Task` class.
    Additional parameters can be specified in the ``PLOT`` dictionary. Currently:

    * **continuous** (bool): whether the data should be plotted against the trial number (False or NA) or against time (True)
    * **chance_bar** (bool): Whether to draw a red horizontal line at chance level (default: 0.5)
    * **chance_level** (float): The position in the y-axis at which the ``chance_bar`` should be drawn
    * **roll_window** (int): The number of trials :class:`~.Roll_Mean` take the average over.

    Attributes:
        pilot (str): The name of our pilot, used to set the identity of our socket, specifically::

            'P_{pilot}'

        infobox (:class:`QtWidgets.QFormLayout`): Box to plot basic task information like trial number, etc.
        info (dict): Widgets in infobox:

            * 'N Trials': :class:`QtWidgets.QLabel`,
            * 'Runtime' : :class:`.Timer`,
            * 'Session' : :class:`QtWidgets.QLabel`,
            * 'Protocol': :class:`QtWidgets.QLabel`,
            * 'Step'    : :class:`QtWidgets.QLabel`

        plot (:class:`pyqtgraph.PlotWidget`): The widget where we draw our plots
        plot_params (dict): A dictionary of plot parameters we receive from the Task class
        data (dict): A dictionary of the data we've received
        plots (dict): The collection of plots we instantiate based on `plot_params`
        node (:class:`.Net_Node`): Our local net node where we listen for data.
        state (str): state of the pilot, used to keep plot synchronized.
    """

    def __init__(self, pilot, x_width=50, parent=None):
        """
        Args:
            pilot (str): The name of our pilot
            x_width (int): How many trials in the past should we plot?
        """
        #super(Plot, self).__init__(QtOpenGL.QGLFormat(QtOpenGL.QGL.SampleBuffers), parent)
        super(Plot, self).__init__()

        self.logger = init_logger(self)

        self.parent = parent
        self.layout = None
        self.infobox = None
        self.n_trials = None
        self.session_trials = 0
        self.info = {}
        self.plot = None
        self.xrange = None
        self.plot_params = {}
        self.data = {} # Keep a dict of the data we are keeping track of, will be instantiated on start
        self.plots = {}
        self.state = "IDLE"
        self.continuous = False
        self.last_time = 0
        self.video = None
        self.videos = []

        self.invoker = get_invoker()

        # The name of our pilot, used to listen for events
        self.pilot = pilot

        # Set initial x-value, will update when data starts coming in
        self.x_width = x_width
        self.last_trial = self.x_width

        # Inits the basic widget settings
        self.init_plots()

        ## Station
        # Start the listener, subscribes to terminal_networking that will broadcast data
        self.listens = {
            'START' : self.l_start, # Receiving a new task
            'DATA' : self.l_data, # Receiving a new datapoint
            'CONTINUOUS': self.l_data,
            'STOP' : self.l_stop,
            'PARAM': self.l_param, # changing some param
            'STATE': self.l_state
        }

        self.node = Net_Node(id='P_{}'.format(self.pilot),
                             upstream="T",
                             port=prefs.get('MSGPORT'),
                             listens=self.listens,
                             instance=True)


    @gui_event
    def init_plots(self):
        """
        Make pre-task GUI objects and set basic visual parameters of `self.plot`
        """

        # This is called to make the basic plot window,
        # each task started should then send us params to populate afterwards
        #self.getPlotItem().hideAxis('bottom')

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.setContentsMargins(2,2,2,2)
        self.setLayout(self.layout)

        # A little infobox to keep track of running time, trials, etc.
        self.infobox = QtWidgets.QFormLayout()
        self.n_trials = count()
        self.session_trials = 0
        self.info = {
            'N Trials': QtWidgets.QLabel(),
            'Runtime' : Timer(),
            'Session' : QtWidgets.QLabel(),
            'Protocol': QtWidgets.QLabel(),
            'Step'    : QtWidgets.QLabel()
        }
        for k, v in self.info.items():

            self.infobox.addRow(k, v)

        #self.infobox.setS


        self.layout.addLayout(self.infobox, 2)

        # The plot that we own :)
        self.plot = pg.PlotWidget()
        self.plot.setContentsMargins(0,0,0,0)

        self.layout.addWidget(self.plot, 8)

        self.xrange = range(self.last_trial - self.x_width + 1, self.last_trial + 1)
        self.plot.setXRange(self.xrange[0], self.xrange[-1])

        self.plot.getPlotItem().hideAxis('left')
        self.plot.setBackground(None)
        self.plot.getPlotItem().getAxis('bottom').setPen({'color':'k'})
        self.plot.getPlotItem().getAxis('bottom').setTickFont('FreeMono')
        self.plot.setXRange(self.xrange[0], self.xrange[1])
        self.plot.enableAutoRange(y=True)
        # self.plot
        # self.plot.setYRange(0, 1)

    @gui_event
    def l_start(self, value):
        """
        Starting a task, initialize task-specific plot objects described in the
        :attr:`.Task.PLOT` attribute.

        Matches the data field name (keys of :attr:`.Task.PLOT` ) to the plot object
        that represents it, eg, to make the standard nafc plot::

            {'target'   : 'point',
             'response' : 'segment',
             'correct'  : 'rollmean'}

        Args:
            value (dict): The same parameter dictionary sent by :meth:`.Terminal.toggle_start`, including

                * current_trial
                * step
                * session
                * step_name
                * task_type
        """

        if self.state in ("RUNNING", "INITIALIZING"):
            return

        self.state = "INITIALIZING"

        # We're sent a task dict, we extract the plot params and send them to the plot object
        self.plot_params = tasks.TASK_LIST[value['task_type']].PLOT

        if 'continuous' in self.plot_params.keys():
            if self.plot_params['continuous']:
                self.continuous = True
            else:
                self.continuous = False
        else:
            self.continuous = False



        # set infobox stuff
        self.n_trials = count()
        self.session_trials = 0
        self.info['N Trials'].setText(str(value['current_trial']))
        self.info['Runtime'].start_timer()
        self.info['Step'].setText(str(value['step']))
        self.info['Session'].setText(str(value['session']))
        self.info['Protocol'].setText(value['step_name'])


        # TODO: Make this more general, make cases for each non-'data' key
        try:
            if self.plot_params['chance_bar']:
                if self.plot_params['chance_level']:
                    try:
                        chance_level = float(self.plot_params['chance_level'])
                    except ValueError:
                        chance_level = 0.5
                        # TODO: Log this.

                    self.plot.getPlotItem().addLine(y=chance_level, pen=(255, 0, 0))

                else:
                    self.plot.getPlotItem().addLine(y=0.5, pen=(255, 0, 0))
        except KeyError:
            # No big deal, chance bar wasn't set
            pass

        # Make plot items for each data type
        for data, plot in self.plot_params['data'].items():
            # TODO: Better way of doing params for plots, redo when params are refactored
            if plot == 'rollmean' and 'roll_window' in self.plot_params.keys():
                self.plots[data] = Roll_Mean(winsize=self.plot_params['roll_window'])
                self.plot.addItem(self.plots[data])
                self.data[data] = np.zeros((0,2), dtype=np.float)
            else:
                self.plots[data] = PLOT_LIST[plot](continuous=self.continuous)
                self.plot.addItem(self.plots[data])
                self.data[data] = np.zeros((0,2), dtype=np.float)

        if 'video' in self.plot_params.keys():
            self.videos = self.plot_params['video']
            self.video = Video(self.plot_params['video'])
            #self.video.start()


        self.state = 'RUNNING'



    @gui_event
    def l_data(self, value):
        """
        Receive some data, if we were told to plot it, stash the data
        and update the assigned plot.

        Args:
            value (dict): Value field of a data message sent during a task.
        """
        if self.state == "INITIALIZING":
            return

        #pdb.set_trace()
        if 'trial_num' in value.keys():
            v = value.pop('trial_num')
            if v >= self.last_trial:
                self.session_trials = next(self.n_trials)
            elif v < self.last_trial:
                self.logger.exception('Shouldnt be going back in time!')
            self.last_trial = v
            # self.last_trial = v
            self.info['N Trials'].setText("{}/{}".format(self.session_trials, v))
            if not self.continuous:
                self.xrange = range(v - self.x_width + 1, v + 1)
                self.plot.setXRange(self.xrange[0], self.xrange[-1])


        if 't' in value.keys():
            self.last_time = value.pop('t')
            if self.continuous:
                self.plot.setXRange(self.last_time-self.x_width, self.last_time+1)


        if self.continuous:
            x_val = self.last_time
        else:
            x_val = self.last_trial

        for k, v in value.items():
            if k in self.data.keys():
                self.data[k] = np.vstack((self.data[k], (x_val, v)))
                # gui_event_fn(self.plots[k].update, *(self.data[k],))
                self.plots[k].update(self.data[k])
            elif k in self.videos:
                self.video.update_frame(k, v)



    @gui_event
    def l_stop(self, value):
        """
        Clean up the plot objects.

        Args:
            value (dict): if "graduation" is a key, don't stop the timer.
        """
        self.data = {}
        self.plots = {}
        self.plot.clear()
        try:
            if isinstance(value, str) or ('graduation' not in value.keys()):
                self.info['Runtime'].stop_timer()
        except:
            self.info['Runtime'].stop_timer()



        self.info['N Trials'].setText('')
        self.info['Step'].setText('')
        self.info['Session'].setText('')
        self.info['Protocol'].setText('')

        if self.video is not None:
            self.video.release()
            self.video.close()
            del self.video
            del self.videos
            self.video = None
            self.videos = []

        self.state = 'IDLE'

    def l_param(self, value):
        """
        Warning:
            Not implemented

        Args:
            value:
        """
        pass

    def l_state(self, value):
        """
        Pilot letting us know its state has changed. Mostly for the case where
        we think we're running but the pi doesn't.

        Args:
            value (:attr:`.Pilot.state`): the state of our pilot

        """

        if (value in ('STOPPING', 'IDLE')) and self.state == 'RUNNING':
            #self.l_stop({})
            pass





###################################
# Curve subclasses
class Point(pg.PlotDataItem):
    """
    A simple point.

    Attributes:
        brush (:class:`QtWidgets.QBrush`)
        pen (:class:`QtWidgets.QPen`)
    """

    def __init__(self, color=(0,0,0), size=5, **kwargs):
        """
        Args:
            color (tuple): RGB color of points
            size (int): width in px.
        """
        super(Point, self).__init__()

        self.continuous = False
        if 'continuous' in kwargs.keys():
            self.continuous = kwargs['continuous']

        self.brush = pg.mkBrush(color)
        self.pen   = pg.mkPen(color, width=size)
        self.size  = size

    def update(self, data):
        """
        Args:
            data (:class:`numpy.ndarray`): an x_width x 2 array where
                column 0 is trial number and column 1 is the value,
                where value can be "L", "C", "R" or a float.
        """
        # data should come in as an n x 2 array,
        # 0th column - trial number (x), 1st - (y) value

        data[data=="R"] = 1
        data[data=="C"] = 0.5
        data[data=="L"] = 0
        data = data.astype(np.float)

        self.scatter.setData(x=data[...,0], y=data[...,1], size=self.size,
                             brush=self.brush, symbol='o', pen=self.pen)

class Line(pg.PlotDataItem):
    """
    A simple line
    """

    def __init__(self, color=(0,0,0), size=1, **kwargs):
        super(Line, self).__init__(**kwargs)

        self.brush = pg.mkBrush(color)
        self.pen = pg.mkPen(color, width=size)
        self.size = size

    def update(self, data):
        data[data=="R"] = 1
        data[data=="L"] = 0
        data[data=="C"] = 0.5
        data = data.astype(np.float)

        self.curve.setData(data[...,0], data[...,1])



class Segment(pg.PlotDataItem):
    """
    A line segment that draws from 0.5 to some endpoint.
    """
    def __init__(self, **kwargs):
        # type: () -> None
        super(Segment, self).__init__(**kwargs)

    def update(self, data):
        """
        data is doubled and then every other value is set to 0.5,
        then :meth:`~pyqtgraph.PlotDataItem.curve.setData` is used with
        `connect='pairs'` to make line segments.

        Args:
            data (:class:`numpy.ndarray`): an x_width x 2 array where
                column 0 is trial number and column 1 is the value,
                where value can be "L", "C", "R" or a float.
        """
        # data should come in as an n x 2 array,
        # 0th column - trial number (x), 1st - (y) value
        data[data=="R"] = 1
        data[data=="L"] = 0
        data[data=="C"] = 0.5
        data = data.astype(np.float)

        xs = np.repeat(data[...,0],2)
        ys = np.repeat(data[...,1],2)
        ys[::2] = 0.5

        self.curve.setData(xs, ys, connect='pairs', pen='k')


class Roll_Mean(pg.PlotDataItem):
    """
    Shaded area underneath a rolling average.

    Typically used as a rolling mean of corrects, so area above and below 0.5 is drawn.
    """
    def __init__(self, winsize=10, **kwargs):
        # type: (int) -> None
        """
        Args:
            winsize (int): number of trials in the past to take a rolling mean of
        """
        super(Roll_Mean, self).__init__()

        self.winsize = winsize

        self.setFillLevel(0.5)

        self.series = pd.Series()

        self.brush = pg.mkBrush((0,0,0,100))
        self.setBrush(self.brush)

    def update(self, data):
        """
        Args:
            data (:class:`numpy.ndarray`): an x_width x 2 array where
                column 0 is trial number and column 1 is the value.
        """
        # data should come in as an n x 2 array,
        # 0th column - trial number (x), 1st - (y) value
        data = data.astype(np.float)

        self.series = pd.Series(data[...,1])
        ys = self.series.rolling(self.winsize, min_periods=0).mean().to_numpy()

        #print(ys)

        self.curve.setData(data[...,0], ys, fillLevel=0.5)

class Shaded(pg.PlotDataItem):
    """
    Shaded area for a continuous plot
    """

    def __init__(self, **kwargs):
        super(Shaded, self).__init__()

        #self.dur = float(dur) # duration of time to display points in seconds
        self.setFillLevel(0)
        self.series = pd.Series()

        self.getBoundingParents()


        self.brush = pg.mkBrush((0,0,0,100))
        self.setBrush(self.brush)

        self.max_num = 0


    def update(self, data):
        """
        Args:
            data (:class:`numpy.ndarray`): an x_width x 2 array where
                column 0 is time and column 1 is the value.
        """
        # data should come in as an n x 2 array,
        # 0th column - trial number (x), 1st - (y) value
        data = data.astype(np.float)

        self.max_num = float(np.abs(np.max(data[:,1])))

        if self.max_num > 1.0:
            data[:,1] = (data[:,1]/(self.max_num*2.0))+0.5
        #print(ys)

        self.curve.setData(data[...,0], data[...,1], fillLevel=0)




class Timer(QtWidgets.QLabel):
    """
    A simple timer that counts... time...

    Uses a :class:`QtCore.QTimer` connected to :meth:`.Timer.update_time` .
    """
    def __init__(self):
        # type: () -> None
        super(Timer, self).__init__()

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_time)

        self.start_time = None

    def start_timer(self, update_interval=1000):
        """
        Args:
            update_interval (float): How often (in ms) the timer should be updated.
        """
        self.start_time = time()
        self.timer.start(update_interval)

    def stop_timer(self):
        """
        you can read the sign ya punk
        """
        self.timer.stop()
        self.setText("")

    def update_time(self):
        """
        Called every (update_interval) milliseconds to set the text of the timer.

        """
        secs_elapsed = int(np.floor(time()-self.start_time))
        self.setText("{:02d}:{:02d}:{:02d}".format(int(secs_elapsed/3600), int((secs_elapsed/60))%60, secs_elapsed%60))


class Video(QtWidgets.QWidget):
    def __init__(self, videos, fps=None):
        """
        Display Video data as it is collected.

        Uses the :class:`ImageItem_TimedUpdate` class to do timed frame updates.

        Args:
            videos (list, tuple): Names of video streams that will be displayed
            fps (int): if None, draw according to ``prefs.get('DRAWFPS')``. Otherwise frequency of widget update

        Attributes:
            videos (list, tuple): Names of video streams that will be displayed
            fps (int): if None, draw according to ``prefs.get('DRAWFPS')``. Otherwise frequency of widget update
            ifps (int): 1/fps, duration of frame in s
            qs (dict): Dictionary of :class:`~queue.Queue`s in which frames will be dumped
            quitting (:class:`threading.Event`): Signal to quit drawing
            update_thread (:class:`threading.Thread`): Thread with target=:meth:`~.Video._update_frame`
            layout (:class:`PySide2.QtWidgets.QGridLayout`): Widget layout
            vid_widgets (dict): dict containing widgets for each of the individual video streams.
        """
        super(Video, self).__init__()

        self.videos = videos

        if fps is None:
            if prefs.get( 'DRAWFPS'):
                self.fps = prefs.get('DRAWFPS')
            else:
                self.fps = 10
        else:
            self.fps = fps

        self.ifps = 1.0/self.fps

        self.layout = None
        self.vid_widgets = {}


        #self.q = Queue(maxsize=1)
        self.qs = {}
        self.quitting = Event()
        self.quitting.clear()


        self.init_gui()

        self.update_thread = Thread(target=self._update_frame)
        self.update_thread.setDaemon(True)
        self.update_thread.start()

    def init_gui(self):
        self.layout = QtWidgets.QGridLayout()
        self.vid_widgets = {}


        for i, vid in enumerate(self.videos):
            vid_label = QtWidgets.QLabel(vid)

            # https://github.com/pyqtgraph/pyqtgraph/blob/3d3d0a24590a59097b6906d34b7a43d54305368d/examples/VideoSpeedTest.py#L51
            graphicsView= pg.GraphicsView(self)
            vb = pg.ViewBox()
            graphicsView.setCentralItem(vb)
            vb.setAspectLocked()
            #img = pg.ImageItem()
            img = ImageItem_TimedUpdate()
            vb.addItem(img)

            self.vid_widgets[vid] = (graphicsView, vb, img)

            # 3 videos in a row
            row = np.floor(i/3.)*2
            col = i%3

            self.layout.addWidget(vid_label, row,col, 1,1)
            self.layout.addWidget(self.vid_widgets[vid][0],row+1,col,5,1)

            # make queue for vid
            self.qs[vid] = Queue(maxsize=1)



        self.setLayout(self.layout)
        self.resize(600,700)
        self.show()

    def _update_frame(self):
        """
        Pulls frames from :attr:`.Video.qs` and feeds them to the video widgets.

        Internal method, run in thread.
        """
        last_time = 0
        this_time = 0
        while not self.quitting.is_set():

            for vid, q in self.qs.items():
                data = None
                try:
                    data = q.get_nowait()
                    self.vid_widgets[vid][2].setImage(data)

                except Empty:
                    pass
                except KeyError:
                    pass

            this_time = time()
            sleep(max(self.ifps-(this_time-last_time), 0))
            last_time = this_time





    def update_frame(self, video, data):
        """
        Put a frame for a video stream into its queue.

        If there is a waiting frame, pull it from the queue first -- it's old now.

        Args:
            video (str): name of video stream
            data (:class:`numpy.ndarray`): video frame
        """
        #pdb.set_trace()
        # cur_time = time()

        try:
            # if there's a waiting frame, it's old now so pull it.
            _ = self.qs[video].get_nowait()
        except Empty:
            pass

        try:
            # put the new frame in there.
            self.qs[video].put_nowait(data)
        except Full:
            return
        except KeyError:
            return

    def release(self):
        self.quitting.set()

#
# class VideoCV(mp.Process):
#     def __init__(self, videos, fps=30, parent=None):
#         super(VideoCV, self).__init__()
#         self.videos = videos
#
#         self.last_update = 0
#         self.fps = fps
#         self.ifps = 1.0/fps
#
#
#         #self.q = Queue(maxsize=1)
#         self.qs = {}
#         for vid in self.videos:
#             self.qs[vid] = mp.Queue(maxsize=1)
#
#         self.positions = {}
#         n_rows = 0
#         n_cols = 0
#         for i, vid in enumerate(sorted(self.videos)):
#             # 3 videos to a row
#             row = np.floor(i/3.)*2
#             col = i%3
#             if row>n_rows:
#                 n_rows = row
#             if col>n_cols:
#                 n_cols = col
#             self.positions[vid] = (row, col)
#
#         self.n_rows = n_rows+1
#         self.n_cols = n_cols+1
#
#
#
#         # computed as we receive images
#         self.sizes = {}
#         self.resize_factors = {}
#
#         self.quitting = mp.Event()
#         self.quitting.clear()
#
#     def run(self):
#
#         win = cv2.namedWindow('vid', cv2.WINDOW_NORMAL)
#         max_width = 0
#         max_height = 0
#
#         img_array = None
#         while not self.quitting.is_set():
#             pdb.set_trace()
#             for vid, q in self.qs.items():
#                 try:
#                     data = q.get_nowait()
#                 except Empty:
#                     continue
#
#                 if vid not in self.sizes.keys():
#                     self.sizes[vid] = (data.shape[0], data.shape[1])
#                     max_width, max_height = self.calc_resize()
#                     img_array = np.zeros((max_height*self.n_rows, max_width*self.n_cols))
#
#                 data = cv2.resize(data, self.resize_factors[vid])
#                 top = self.positions[vid][0] * max_height
#                 left = self.positions[vid][1] * max_width
#                 img_array[top:top+data.shape[0], left:left+data.shape[1]] = data
#
#             cv2.imshow('vid', img_array)
#             cv2.waitKey(0)
#
#
#
#     def calc_resize(self):
#         max_width = 0
#         max_height = 0
#         for vid, size in self.sizes:
#             if size[1]>max_width:
#                 # set both so we don't split
#                 max_width = size[1]
#                 max_height = size[0]
#             elif size[0]>max_height:
#                 max_width = size[1]
#                 max_height=size[0]
#
#         for vid, size in self.sizes:
#             self.resize_factors[vid] = (float(max_width)/size[0], float(max_height)/size[1])
#
#         return max_width, max_height
#
#
#     def update_frame(self, video, data):
#         #pdb.set_trace()
#         # cur_time = time()
#
#         try:
#             # if there's a waiting frame, it's old now so pull it.
#             _ = self.qs[video].get_nowait()
#         except Empty:
#             pass
#
#         try:
#             # put the new frame in there.
#             self.qs[video].put_nowait(data)
#         except Full:
#             return
#         except KeyError:
#             return
#         # if (cur_time-self.last_update)>self.ifps:
#         #     try:
#         #         self.vid_widgets[video].setImage(data)
#         #         #self.vid_widgets[video].update()
#         #     except KeyError:
#         #         return
#         #     self.last_update = cur_time
#             #self.update()
#             #self.app.processEvents()
#
#     def close(self):
#         self.quitting.set()







# class Highlight():
#     # TODO Implement me
#     def __init__(self):
#         pass
#
#     pass


class HLine(QtWidgets.QFrame):
    """
    A Horizontal line.
    """
    def __init__(self):
        # type: () -> None
        super(HLine, self).__init__()
        self.setFrameShape(QtWidgets.QFrame.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Sunken)

VIDEO_TIMER = None

class ImageItem_TimedUpdate(pg.ImageItem):
    """
    Reclass of :class:`pyqtgraph.ImageItem` to update with a fixed fps.

    Rather than calling :meth:`~pyqtgraph.ImageItem.update` every time a frame is updated,
    call it according to the timer.

    fps is set according to ``prefs.get('DRAWFPS')``, if not available, draw at 10fps

    Attributes:
        timer (:class:`~PySide2.QtCore.QTimer`): Timer held in ``globals()`` that synchronizes frame updates across
            image items


    """

    def __init__(self, *args, **kwargs):
        super(ImageItem_TimedUpdate, self).__init__(*args, **kwargs)

        if globals()['VIDEO_TIMER'] is None:
            globals()['VIDEO_TIMER'] = QtCore.QTimer()


        self.timer = globals()['VIDEO_TIMER']
        self.timer.stop()
        self.timer.timeout.connect(self.update_img)
        if prefs.get( 'DRAWFPS'):
            self.fps = prefs.get('DRAWFPS')
        else:
            self.fps = 10.
        self.timer.start(1./self.fps)




    def setImage(self, image=None, autoLevels=None, **kargs):
        """
        Update the image displayed by this item. For more information on how the image
        is processed before displaying, see :func:`makeARGB <pyqtgraph.makeARGB>`
        =================  =========================================================================
        **Arguments:**
        image              (numpy array) Specifies the image data. May be 2D (width, height) or
                           3D (width, height, RGBa). The array dtype must be integer or floating
                           point of any bit depth. For 3D arrays, the third dimension must
                           be of length 3 (RGB) or 4 (RGBA). See *notes* below.
        autoLevels         (bool) If True, this forces the image to automatically select
                           levels based on the maximum and minimum values in the data.
                           By default, this argument is true unless the levels argument is
                           given.
        lut                (numpy array) The color lookup table to use when displaying the image.
                           See :func:`setLookupTable <pyqtgraph.ImageItem.setLookupTable>`.
        levels             (min, max) The minimum and maximum values to use when rescaling the image
                           data. By default, this will be set to the minimum and maximum values
                           in the image. If the image array has dtype uint8, no rescaling is necessary.
        opacity            (float 0.0-1.0)
        compositionMode    See :func:`setCompositionMode <pyqtgraph.ImageItem.setCompositionMode>`
        border             Sets the pen used when drawing the image border. Default is None.
        autoDownsample     (bool) If True, the image is automatically downsampled to match the
                           screen resolution. This improves performance for large images and
                           reduces aliasing. If autoDownsample is not specified, then ImageItem will
                           choose whether to downsample the image based on its size.
        =================  =========================================================================
        **Notes:**
        For backward compatibility, image data is assumed to be in column-major order (column, row).
        However, most image data is stored in row-major order (row, column) and will need to be
        transposed before calling setImage()::
            imageitem.setImage(imagedata.T)
        This requirement can be changed by calling ``image.setOpts(axisOrder='row-major')`` or
        by changing the ``imageAxisOrder`` :ref:`global configuration option <apiref_config>`.
        """
        #profile = debug.Profiler()

        gotNewData = False
        if image is None:
            if self.image is None:
                return
        else:
            gotNewData = True
            shapeChanged = (self.image is None or image.shape != self.image.shape)
            image = image.view(np.ndarray)
            if self.image is None or image.dtype != self.image.dtype:
                self._effectiveLut = None
            self.image = image
            if self.image.shape[0] > 2 ** 15 - 1 or self.image.shape[1] > 2 ** 15 - 1:
                if 'autoDownsample' not in kargs:
                    kargs['autoDownsample'] = True
            if shapeChanged:
                self.prepareGeometryChange()
                self.informViewBoundsChanged()

        #profile()

        if autoLevels is None:
            if 'levels' in kargs:
                autoLevels = False
            else:
                autoLevels = True
        if autoLevels:
            img = self.image
            while img.size > 2 ** 16:
                img = img[::2, ::2]
            mn, mx = np.nanmin(img), np.nanmax(img)
            # mn and mx can still be NaN if the data is all-NaN
            if mn == mx or np.isnan(mn) or np.isnan(mx):
                mn = 0
                mx = 255
            kargs['levels'] = [mn, mx]


        self.setOpts(update=False, **kargs)

        self.qimage = None

        if gotNewData:
            self.sigImageChanged.emit()

    def update_img(self):
        """
        Call :meth:`~ImageItem_TimedUpdate.update`
        """
        self.update()

    def __del__(self):
        super(ImageItem_TimedUpdate,self).__del__()
        self.timer.stop()




PLOT_LIST = {
    'point':Point,
    'segment':Segment,
    'rollmean':Roll_Mean,
    'shaded':Shaded,
    'line': Line
    # 'highlight':Highlight
}
"""
A dictionary connecting plot keys to objects.

TODO:
    Just reference the plot objects.
"""


"""
Classes to play sounds.

Each sound inherits a base type depending on `prefs.get('AUDIOSERVER')`

* `prefs.get('AUDIOSERVER') == 'jack'` : :class:`.Jack_Sound`
* `prefs.get('AUDIOSERVER') == 'pyo'` : :class:`.Pyo_Sound`

To avoid unnecessary dependencies, `Jack_Sound` is not defined if AUDIOSERVER is `'pyo'`
and vice versa.

TODO:
    Implement sound level and filter calibration


"""

# Re: The organization of this module
# We balance a few things:
# 1) using two sound servers with very different approaches to
# delivering sounds, and
# 2) having a similar API so other modules can query sound properties
# while still being agnostic to the sound server.
# 3) not have our classes split into a ton of Pyo_Tone, Jack_Tone
# copies so they have their parameters and behavior drift apart
#
# So, We have base classes, but they can't encapsulate all the
# behavior for making sounds, so use an init_audio() method that
# creates sound conditional on the type of audio server.

# TODO: Be a whole lot more robust about handling different numbers of channels

import os
import sys
from time import sleep
from scipy.io import wavfile
from scipy.signal import resample
import numpy as np
import threading
import logging
from itertools import cycle
import warnings
if sys.version_info >= (3,0):
    from queue import Empty, Full
else:
    from Queue import Empty, Full


from autopilot import prefs
from autopilot.core.loggers import init_logger

# switch behavior based on audio server type
try:
    if isinstance(prefs.get('AUDIOSERVER'), str):
        server_type = prefs.get('AUDIOSERVER').lower()
    else:
        server_type = prefs.get('AUDIOSERVER')
except:
#    # TODO: The 'attribute don't exist' type - i think NameError?
    server_type = None



if server_type in ("pyo", "docs"):
    try:
        import pyo
    except ImportError:
        pass

    class Pyo_Sound(object):
        """
        Metaclass for pyo sound objects.

        Note:
            Use of pyo is generally discouraged due to dropout issues and
            the general opacity of the module. As such this object is
            intentionally left undocumented.

        """


        def __init__(self):
            self.PARAMS = None  # list of strings of parameters to be defined
            self.type = None  # string human readable name of sound
            self.duration = None  # duration in ms
            self.amplitude = None
            self.table = None
            self.trigger = None
            self.server_type = 'pyo'


        def play(self):
            self.table.out()

        def table_wrap(self, audio, duration=None):
            """Records a PyoAudio generator into a sound table, returns a
            tableread object which can play the audio with .out()

            Args:
                audio:
                duration:
            """

            if not duration:
                duration = self.duration

            # Duration is in ms, so divide by 1000
            # See https://groups.google.com/forum/#!topic/pyo-discuss/N-pan7wPF-o
            # TODO: Get chnls to be responsive to NCHANNELS in prefs. hardcoded for now
            tab = pyo.NewTable(length=(float(duration) / 1000),
                               chnls=prefs.get('NCHANNELS'))  # Prefs should always be declared in the global namespace
            tabrec = pyo.TableRec(audio, table=tab, fadetime=0.005).play()
            sleep((float(duration) / 1000))
            self.table = pyo.TableRead(tab, freq=tab.getRate(), loop=0)

        def set_trigger(self, trig_fn):
            """
            Args:
                trig_fn:
            """
            # Using table triggers, call trig_fn when table finishes playing
            self.trigger = pyo.TrigFunc(self.table['trig'], trig_fn)


if server_type in ("jack", "docs", True):
    if server_type is True:
        server_type = "jack"
    from autopilot.stim.sound import jackclient

    class Jack_Sound(object):
        """
        Base class for sounds that use the :class:`~.jackclient.JackClient` audio
        server.

        Attributes:
            PARAMS (list): List of strings of parameters that need to be defined for this sound
            type (str): Human readable name of sound type
            duration (float): Duration of sound in ms
            amplitude (float): Amplitude of sound as proportion of 1 (eg 0.5 is half amplitude)
            table (:class:`numpy.ndarray`): A Numpy array of samples
            chunks (list): :attr:`~.Jack_Sound.table` split up into chunks of :data:`~.jackclient.BLOCKSIZE`
            trigger (callable): A function that is called when the sound completes
            nsamples (int): Number of samples in the sound
            padded (bool): Whether the sound had to be padded with zeros when split into chunks (ie. sound duration was not a multiple of BLOCKSIZE).
            fs (int): sampling rate of client from :data:`.jackclient.FS`
            blocksize (int): blocksize of client from :data:`.jackclient.BLOCKSIZE`
            server (:class:`~.jackclient.Jack_Client`): Current Jack Client
            q (:class:`multiprocessing.Queue`): Audio Buffer queue from :data:`.jackclient.QUEUE`
            q_lock (:class:`multiprocessing.Lock`): Audio Buffer lock from :data:`.jackclient.Q_LOCK`
            play_evt (:class:`multiprocessing.Event`): play event from :data:`.jackclient.PLAY`
            stop_evt (:class:`multiprocessing.Event`): stop event from :data:`.jackclient.STOP`
            buffered (bool): has this sound been dumped into the :attr:`~.Jack_Sound.q` ?
            buffered_continuous (bool): Has the sound been dumped into the :attr:`~.Jack_Sound.continuous_q`?

        """

        PARAMS = []
        """
        list:  list of strings of parameters to be defined
        """

        type = None
        """
        str: string human readable name of sound
        """

        server_type = 'jack'
        """
        str: type of server, always 'jack' for `Jack_Sound` s.
        """

        def __init__(self):
            self.duration = None  # duration in ms
            self.amplitude = None
            self.table = None  # numpy array of samples
            self.chunks = None  # table split into a list of chunks
            self.trigger = None
            self.nsamples = None
            self.padded = False # whether or not the sound was padded with zeros when chunked
            self.continuous = False


            self.fs = jackclient.FS
            self.blocksize = jackclient.BLOCKSIZE
            self.server = jackclient.SERVER
            self.q = jackclient.QUEUE
            self.q_lock = jackclient.Q_LOCK
            self.play_evt = jackclient.PLAY
            self.stop_evt = jackclient.STOP
            self.continuous_flag = jackclient.CONTINUOUS
            self.continuous_q = jackclient.CONTINUOUS_QUEUE
            self.continuous_loop = jackclient.CONTINUOUS_LOOP
            self.quitting = threading.Event()

            self.initialized = False
            self.buffered = False
            self.buffered_continuous = False

            self.logger = init_logger(self)


        def chunk(self, pad=True):
            """
            Split our `table` up into a list of :attr:`.Jack_Sound.blocksize` chunks.

            Args:
                pad (bool): If the sound is not evenly divisible into chunks, pad with zeros (True, default), otherwise jackclient will pad with its continuous sound
            """
            # break sound into chunks

            sound = self.table.astype(np.float32)
            sound_list = [sound[i:i+self.blocksize] for i in range(0, sound.shape[0], self.blocksize)]

            if (sound_list[-1].shape[0] < self.blocksize) and pad:
                sound_list[-1] = np.pad(sound_list[-1],
                                        (0, self.blocksize-sound_list[-1].shape[0]),
                                        'constant')
                self.padded = True
            else:
                self.padded = False

            self.chunks = sound_list

        def set_trigger(self, trig_fn):
            """
            Set a trigger function to be called when the :attr:`~.Jack_Sound.stop_evt` is set.

            Args:
                trig_fn (callable): Some callable
            """
            if callable(trig_fn):
                self.trigger = trig_fn
            else:
                Exception('trigger must be callable')

        def wait_trigger(self):
            """
            Wait for the stop_evt trigger to be set for at least a second after
            the sound should have ended.

            Call the trigger when the event is set.
            """
            # wait for our duration plus a second at most.
            self.stop_evt.wait((self.duration+1000)/1000.)
            # if the sound actually stopped...
            if self.stop_evt.is_set():
                self.trigger()

        def get_nsamples(self):
            """
            given our fs and duration, how many samples do we need?

            literally::

                np.ceil((self.duration/1000.)*self.fs).astype(np.int)

            """
            self.nsamples = np.ceil((self.duration/1000.)*self.fs).astype(np.int)

        def quantize_duration(self, ceiling=True):
            """
            Extend or shorten a sound so that it is a multiple of :data:`.jackclient.BLOCKSIZE`

            Args:
                ceiling (bool): If true, extend duration, otherwise decrease duration.
            """

            # get remainder of samples
            self.get_nsamples()
            remainder = self.nsamples % self.blocksize

            if remainder == 0:
                return

            # get target number of samples
            # get target n blocks and multiply by blocksize
            if ceiling:
                target_samples = np.ceil(float(self.nsamples)/self.blocksize)*self.blocksize
            else:
                target_samples = np.floor(float(self.nsamples)/self.blocksize)*self.blocksize

            # get new duration
            self.duration = (target_samples/self.fs)*1000.

            # refresh nsamples
            self.get_nsamples()





        def buffer(self):
            """
            Dump chunks into the sound queue.
            """

            if hasattr(self, 'path'):
                self.logger.debug('BUFFERING SOUND {}'.format(self.path))

            if not self.initialized and not self.table:
                try:
                    self.init_sound()
                    self.initialized = True
                except:
                    pass
                    #TODO: Log this, better error handling here

            if not self.chunks:
                self.chunk()

            with self.q_lock:
                # empty queue
                # FIXME: Testing whether this is where we get held up on the 'fail after sound play' bug
                # n_gets = 0
                while not self.q.empty():
                    try:
                        _ = self.q.get_nowait()
                    except Empty:
                        # normal, get until it's empty
                        break
                    # n_gets += 1
                    # if n_gets > 100000:
                    #     break
                for frame in self.chunks:
                    self.q.put_nowait(frame)
                # The jack server looks for a None object to clear the play flag
                self.q.put_nowait(None)
                self.buffered = True

        def buffer_continuous(self):
            """
            Dump chunks into the continuous sound queue for looping.

            Continuous shoulds should always have full frames -
            ie. the number of samples in a sound should be a multiple of :data:`.jackclient.BLOCKSIZE`.

            This method will call :meth:`.quantize_duration` to force duration such that the sound has full frames.

            An exception will be raised if the sound has been padded.



            Returns:

            """

            # FIXME: Initialized should be more flexible,
            # for now just deleting whatever init happened because
            # continuous sounds are in development
            self.table = None
            self.initialized = False

            if not self.initialized and not self.table:
                # try:
                self.quantize_duration()
                self.init_sound()
                self.initialized = True
                # except:
                #     pass
                    #TODO: Log this, better error handling here

            if not self.chunks:
                self.chunk()

            # continous sounds should not have any padding - see docstring
            if self.padded:
                raise Exception("Continuous sounds cannot have padded chunks - sounds need to have n_samples % blocksize == 0")

            # empty queue
            while not self.continuous_q.empty():
                try:
                    _ = self.continuous_q.get_nowait()
                except Empty:
                    # normal, get until it's empty
                    break

            # load frames into continuous queue
            for frame in self.chunks:
                self.continuous_q.put_nowait(frame)
            # The jack server looks for a None object to clear the play flag
            # self.continuous_q.put_nowait(None)
            self.buffered_continuous = True


        def play(self):
            """
            Play ourselves.

            If we're not buffered, be buffered.

            Otherwise, set the play event and clear the stop event.

            If we have a trigger, set a Thread to wait on it.
            """
            if not self.buffered:
                self.buffer()

            if hasattr(self, 'path'):
                self.logger.debug('PLAYING SOUND {}'.format(self.path))

            self.play_evt.set()
            self.stop_evt.clear()
            self.buffered = False

            if callable(self.trigger):
                threading.Thread(target=self.wait_trigger).start()

        def play_continuous(self, loop=True):
            """
            Play the sound continuously.

            Sound will be paused if another sound has its 'play' method called.

            Currently - only looping is implemented: the full sound is loaded by the jack client and repeated indefinitely.

            In the future, sound generation methods will be refactored as python generators so sounds can be continuously generated and played.

            Args:
                loop (bool): whether the sound will be stored by the jack client and looped (True), or whether the sound will be continuously streamed (False, not implemented)

            Returns:

            todo::

                merge into single play method that changes behavior if continuous or not

            """

            if not loop:
                raise NotImplementedError('Continuous, unlooped streaming has not been implemented yet!')

            # FIXME: Initialized should be more flexible,
            # for now just deleting whatever init happened because
            # continuous sounds are in development
            self.table = None
            self.initialized = False

            self.quantize_duration()
            self.init_sound()
            self.initialized = True
            self.chunk()

            # if not self.buffered_continuous:
            #     self.buffer_continuous()
            self.continuous_cycle = cycle(self.chunks)

            # start the buffering thread
            self.cont_thread = threading.Thread(target=self._buffer_continuous)
            self.cont_thread.setDaemon(True)
            self.cont_thread.start()

            if loop:
                self.continuous_loop.set()
            else:
                self.continuous_loop.clear()


            # tell the sound server that it has a continuous sound now
            self.continuous_flag.set()
            self.continuous = True

        def _buffer_continuous(self):

            # empty queue
            while not self.continuous_q.empty():
                try:
                    _ = self.continuous_q.get_nowait()
                except Empty:
                    # normal, get until it's empty
                    break

            # want to be able to quit if queue remains full for, say, 20 periods
            #wait_time = (self.blocksize/float(self.fs))*20

            while not self.quitting.is_set():
                try:
                    #self.continuous_q.put(self.continuous_cycle.next(), timeout=wait_time)
                    self.continuous_q.put_nowait(next(self.continuous_cycle))
                except Full:
                    pass
            # for chunk in self.chunks:
            #     self.continuous_q.put_nowait(chunk)

        def stop_continuous(self):
            """
            Stop playing a continuous sound

            Should be merged into a general stop method
            """
            if not self.continuous:
                self.logger.warning("stop_continuous called but not a continuous sound!")
                return

            self.quitting.set()
            self.continuous_flag.clear()
            self.continuous_loop.clear()






        def end(self):
            """
            Release any resources held by this sound

            """

            if self.play_evt.is_set():
                self.play_evt.clear()

            if not self.stop_evt.is_set():
                self.stop_evt.set()

            if self.continuous:
                while not self.continuous_q.empty():
                    try:
                        _ = self.continuous_q.get_nowait()
                    except Empty:
                        # normal, get until it's empty
                        break
                self.buffered_continuous = False
                self.continuous_flag.clear()

            self.table = None
            self.initialized = False

        def __del__(self):
            self.end()




else:
    # just importing to query parameters, not play sounds.
    pass





####################
if server_type == "pyo":
    BASE_CLASS = Pyo_Sound
elif server_type == "jack":
    BASE_CLASS = Jack_Sound
else:
    # just importing to query parameters, not play sounds.
    BASE_CLASS = object
    warnings.warn('No Base Sound class specified! sounds will probably not work!!!')


class Tone(BASE_CLASS):
    """The Humble Sine Wave"""

    PARAMS = ['frequency','duration','amplitude']
    type = 'Tone'

    def __init__(self, frequency, duration, amplitude=0.01, **kwargs):
        """
        Args:
            frequency (float): frequency of sin in Hz
            duration (float): duration of the sin in ms
            amplitude (float): amplitude of the sound as a proportion of 1.
            **kwargs: extraneous parameters that might come along with instantiating us
        """
        super(Tone, self).__init__()

        self.frequency = float(frequency)
        self.duration = float(duration)
        self.amplitude = float(amplitude)

        self.init_sound()

    def init_sound(self):
        """
        Create a sine wave table using pyo or numpy, depending on the server type.
        """

        if self.server_type == 'pyo':
            sin = pyo.Sine(self.frequency, mul=self.amplitude)
            self.table = self.table_wrap(sin)
        elif self.server_type == 'jack':
            self.get_nsamples()
            t = np.arange(self.nsamples)
            self.table = (self.amplitude*np.sin(2*np.pi*self.frequency*t/self.fs)).astype(np.float32)
            #self.table = np.column_stack((self.table, self.table))
            self.chunk()

        self.initialized = True

class Noise(BASE_CLASS):
    """White Noise"""

    PARAMS = ['duration','amplitude']
    type='Noise'
    def __init__(self, duration, amplitude=0.01, **kwargs):
        """
        Args:
            duration (float): duration of the noise
            amplitude (float): amplitude of the sound as a proportion of 1.
            **kwargs: extraneous parameters that might come along with instantiating us
        """
        super(Noise, self).__init__()

        self.duration = float(duration)
        self.amplitude = float(amplitude)

        self.init_sound()

    def init_sound(self):
        """
        Create a table of Noise using pyo or numpy, depending on the server_type
        """

        if self.server_type == 'pyo':
            noiser = pyo.Noise(mul=self.amplitude)
            self.table = self.table_wrap(noiser)
        elif self.server_type == 'jack':
            self.get_nsamples()
            # rand generates from 0 to 1, so subtract 0.5, double to get -1 to 1,
            # then multiply by amplitude.
            self.table = (self.amplitude * np.random.uniform(-1,1,self.nsamples)).astype(np.float32)
            self.chunk()

        self.initialized = True

class File(BASE_CLASS):
    """
    A .wav file.

    TODO:
        Generalize this to other audio types if needed.
    """

    PARAMS = ['path', 'amplitude']
    type='File'

    def __init__(self, path, amplitude=0.01, **kwargs):
        """
        Args:
            path (str): Path to a .wav file relative to the `prefs.get('SOUNDDIR')`
            amplitude (float): amplitude of the sound as a proportion of 1.
            **kwargs: extraneous parameters that might come along with instantiating us
        """
        super(File, self).__init__()

        if os.path.exists(path):
            self.path = path
        elif os.path.exists(os.path.join(prefs.get('SOUNDDIR'), path)):
            self.path = os.path.join(prefs.get('SOUNDDIR'), path)
        else:
            Exception('Could not find {} in current directory or sound directory'.format(path))

        self.amplitude = float(amplitude)

        # because files can be v memory intensive, we only load the sound once we're called to buffer them
        # store our initialization status
        self.initialized = False

        #self.init_sound()

    def init_sound(self):
        """
        Load the wavfile with :mod:`scipy.io.wavfile` ,
        converting int to float as needed.

        Create a sound table, resampling sound if needed.


        """

        fs, audio = wavfile.read(self.path)
        if audio.dtype in ['int16', 'int32']:
            audio = int_to_float(audio)

        # load file to sound table
        if self.server_type == 'pyo':
            self.dtable = pyo.DataTable(size=audio.shape[0], chnls=prefs.get('NCHANNELS'), init=audio.tolist())

            # get server to determine sampling rate modification and duration
            server_fs = self.dtable.getServer().getSamplingRate()
            self.duration = float(self.dtable.getSize()) / float(fs)
            self.table = pyo.TableRead(table=self.dtable, freq=float(fs) / server_fs,
                                       loop=False, mul=self.amplitude)

        elif self.server_type == 'jack':
            # attenuate amplitude
            audio = audio*self.amplitude
            self.duration = float(audio.shape[0]) / fs
            # resample to match our audio server's sampling rate
            if fs != self.fs:
                new_samples = self.duration*self.fs
                audio = resample(audio, new_samples)

            self.table = audio

        self.initialized = True


class Speech(File):
    """
    Speech subclass of File sound.

    Example of custom sound class - PARAMS are changed, but nothing else.
    """

    type='Speech'
    PARAMS = ['path', 'amplitude', 'speaker', 'consonant', 'vowel', 'token']
    def __init__(self, path, speaker, consonant, vowel, token, amplitude=0.05, **kwargs):
        """
        Args:
            speaker (str): Which Speaker recorded this speech token?
            consonant (str): Which consonant is in this speech token?
            vowel (str): Which vowel is in this speech token?
            token (int): Which token is this for a given combination of speaker, consonant, and vowel
        """
        super(Speech, self).__init__(path, amplitude, **kwargs)

        self.speaker = speaker
        self.consonant = consonant
        self.vowel = vowel
        self.token = token

        # sound is init'd in the superclass

class Gap(BASE_CLASS):
    """
    A silent sound that does not pad its final chunk -- used for creating precise silent
    gaps in a continuous noise.

    """

    type = "Gap"
    PARAMS = ['duration']

    def __init__(self, duration, **kwargs):
        """
        Args:
            duration (float): duration of gap in ms

        Attributes:
            gap_zero (bool): True if duration is zero, effectively do nothing on play.
        """
        super(Gap, self).__init__()

        self.duration = float(duration)
        self.gap_zero = False

        if self.duration == 0:
            self.gap_zero = True
            self.get_nsamples()
            self.chunks = []
            self.table = np.ndarray((0,),dtype=np.float32)
            self.initialized = True
        else:

            self.init_sound()

    def init_sound(self):
        """
        Create and chunk an array of zeros according to :attr:`.Gap.duration`
        """
        if self.server_type == "pyo":
            raise NotImplementedError("This sound has not been implemented for pyo sound server -- pyo is deprecated, and kept as a skeleton in the case interested programmers want to revive its use")

        # get the number of samples for the sound given our self.duration
        self.get_nsamples()
        self.table = np.zeros((self.nsamples,), dtype=np.float32)

        # chunk without padding -- jackclient will pad with ongoing continuous noise (or silence if none)
        self.chunk(pad=False)

        self.initialized = True

    def chunk(self, pad=False):
        """
        If gap is not duration == 0, call parent ``chunk``.
        Args:
            pad (bool): unused, passed to parent ``chunk``
        """
        if not self.gap_zero:
            super(Gap, self).chunk(pad)
        else:
            self.padded=False


    def buffer(self):
        if not self.gap_zero:
            super(Gap, self).buffer()
        else:
            self.buffered = True

    def play(self):
        if not self.gap_zero:
            super(Gap, self).play()
        else:
            if callable(self.trigger):
                threading.Thread(target=self.wait_trigger).start()













#######################
# Has to be at bottom so fnxns already defined when assigned.
SOUND_LIST = {
    'Tone':Tone,
    'Noise':Noise,
    'File':File,
    'Speech':Speech,
    'speech':Speech,
    'Gap': Gap
}
"""
Sounds must be added to this SOUND_LIST so they can be indexed by the string keys used elsewhere. 
"""

# These parameters are strings not numbers... jonny should do this better
STRING_PARAMS = ['path', 'speaker', 'consonant', 'vowel', 'type']
"""
These parameters should be given string columns rather than float columns.

Bother Jonny to do this better.

v0.3 will be all about doing parameters better.
"""


def int_to_float(audio):
    """
    Convert 16 or 32 bit integer audio to 32 bit float.

    Args:
        audio (:class:`numpy.ndarray`): a numpy array of audio

    Returns:
        :class:`numpy.ndarray`: Audio that has been rescaled and converted to a 32 bit float.
    """
    if audio.dtype == 'int16':
        audio = audio.astype(np.float32)
        audio = audio / (float(2 ** 16) / 2)
    elif audio.dtype == 'int32':
        audio = audio.astype(np.float32)
        audio = audio / (float(2 ** 32) / 2)

    return audio















from autopilot.tasks.task import Task
from autopilot.tasks.nafc import Nafc, Nafc_Gap, Nafc_Gap_Laser
from autopilot.tasks.gonogo import GoNoGo
from autopilot.tasks.parallax import Parallax
from autopilot.tasks.free_water import Free_Water
from autopilot.tasks.graduation import GRAD_LIST
from autopilot.tasks.children import Wheel_Child, Video_Child, Transformer
from autopilot.tasks.test import DLC_Latency, DLC_Hand


TASK_LIST = {'2AFC':Nafc,
             '2AFC_Gap':Nafc_Gap,
             '2AFC_Gap_Laser':Nafc_Gap_Laser,
             'Free Water':Free_Water,
             'GoNoGo': GoNoGo,
             'Parallax': Parallax,
             'Test_DLC_Latency': DLC_Latency,
             'Test_DLC_Hand':DLC_Hand}
"""
Link between string task names used in protocol descriptions and task classes
"""

CHILDREN_LIST = {
    'Wheel Child':Wheel_Child,
    'Video Child':Video_Child,
    'Transformer': Transformer
}
"""
Link between string child names used in protocol descriptions and task classes
"""




"""
For converting between things that are the same thing but have different numbers and shapes
"""

import typing
from enum import Enum, auto
import numpy as np
import colorsys

from autopilot.transform.transforms import Transform


class Rescale(Transform):
    """
    Rescale values from one range to another
    """

    format_in = {'type': (np.ndarray, float, int, tuple, list)}
    format_out = {'type': np.ndarray}

    def __init__(self,
                 in_range: typing.Tuple[float, float] = (0, 1),
                 out_range: typing.Tuple[float, float] = (0, 1),
                 clip = False,
                 *args, **kwargs):

        super(Rescale, self).__init__(*args, **kwargs)

        self.in_range = in_range
        self.out_range = out_range

        self.in_diff = self.in_range[1] - self.in_range[0]
        self.out_diff = self.out_range[1] - self.out_range[0]
        self.ratio = self.out_diff / self.in_diff

        self.clip = clip

    def process(self, input):
        """
        Subtract input minimum, multiple by output/input size ratio, add output minimum
        """
        if isinstance(input, (tuple, list)):
            input = np.array(input)

        input = ((input - self.in_range[0]) * self.ratio) + self.out_range[0]

        if self.clip:
            input = np.clip(input, self.out_range[0], self.out_range[1])
        return input

class Colorspaces(Enum):
    HSV = auto()
    RGB = auto()
    YIQ = auto()
    HLS = auto()



class Color(Transform):
    """
    Convert colors using the colorsys module!!

    .. note::

        All inputs must be scaled (0,1) and all outputs will be (0,1)
    """
    format_in = {'type': tuple}
    format_out = {'type': tuple}

    CONVERSIONS = {}
    CONVERSIONS[Colorspaces.RGB] = {}
    CONVERSIONS[Colorspaces.YIQ] = {}
    CONVERSIONS[Colorspaces.HLS] = {}
    CONVERSIONS[Colorspaces.HSV] = {}
    CONVERSIONS[Colorspaces.RGB][Colorspaces.YIQ] = colorsys.rgb_to_yiq
    CONVERSIONS[Colorspaces.YIQ][Colorspaces.RGB] = colorsys.yiq_to_rgb
    CONVERSIONS[Colorspaces.RGB][Colorspaces.HLS] = colorsys.rgb_to_hls
    CONVERSIONS[Colorspaces.HLS][Colorspaces.RGB] = colorsys.hls_to_rgb
    CONVERSIONS[Colorspaces.RGB][Colorspaces.HSV] = colorsys.rgb_to_hsv
    CONVERSIONS[Colorspaces.HSV][Colorspaces.RGB] = colorsys.hsv_to_rgb

    def __init__(self,
                 convert_from: Colorspaces = Colorspaces.HSV,
                 convert_to: Colorspaces = Colorspaces.RGB,
                 output_scale = 255, *args, **kwargs):
        super(Color, self).__init__(*args, **kwargs)

        self.convert_from = convert_from
        self.convert_to = convert_to

        # select processing function based on colors given
        self.process_fn = self.CONVERSIONS[self.convert_from][self.convert_to]

    def process(self, input, *args):
        if len(args)>0:
            return np.array(self.process_fn(input, args[0], args[1]))
        else:
            return np.array(self.process_fn(*input))







from collections import OrderedDict as odict
import tables
import itertools
import random
import datetime

import autopilot.hardware.gpio

from autopilot.tasks.task import Task

TASK = 'Free_water'

class Free_Water(Task):
    """
    Randomly light up one of the ports, then dispense water when the subject pokes there

    Two stages:

    * waiting for response, and
    * reporting the response afterwards

    Attributes:
        target ('L', 'C', 'R'): The correct port
        trial_counter (:class:`itertools.count`): Counts trials starting from current_trial specified as argument
        triggers (dict): Dictionary mapping triggered pins to callable methods.
        num_stages (int): number of stages in task (2)
        stages (:class:`itertools.cycle`): iterator to cycle indefinitely through task stages.
    """

    STAGE_NAMES = ["water", "response"]

    # Params
    PARAMS = odict()
    PARAMS['reward'] = {'tag':'Reward Duration (ms)',
                        'type':'int'}
    PARAMS['allow_repeat'] = {'tag':'Allow Repeated Ports?',
                              'type':'bool'}

    # Returned Data
    DATA = {
        'trial_num': {'type':'i32'},
        'target': {'type':'S1', 'plot':'target'},
        'timestamp': {'type':'S26'}, # only one timestamp, since next trial instant
    }

    # TODO: This should be generated from DATA above. Perhaps parsimoniously by using tables types rather than string descriptors
    class TrialData(tables.IsDescription):
        trial_num = tables.Int32Col()
        target    = tables.StringCol(1)
        timestamp = tables.StringCol(26)

    HARDWARE = {
        'POKES':{
            'L': autopilot.hardware.gpio.Digital_In,
            'C': autopilot.hardware.gpio.Digital_In,
            'R': autopilot.hardware.gpio.Digital_In
        },
        'LEDS':{
            # TODO: use LEDs, RGB vs. white LED option in init
            'L': autopilot.hardware.gpio.LED_RGB,
            'C': autopilot.hardware.gpio.LED_RGB,
            'R': autopilot.hardware.gpio.LED_RGB
        },
        'PORTS':{
            'L': autopilot.hardware.gpio.Solenoid,
            'C': autopilot.hardware.gpio.Solenoid,
            'R': autopilot.hardware.gpio.Solenoid
        }
    }

    # Plot parameters
    PLOT = {
        'data': {
            'target': 'point'
        }
    }

    def __init__(self, stage_block=None, current_trial=0,
                 reward=50, allow_repeat=False, **kwargs):
        """
        Args:
            stage_block (:class:`threading.Event`): used to signal to the carrying Pilot that the current trial stage is over
            current_trial (int): If not zero, initial number of `trial_counter`
            reward (int): ms to open solenoids
            allow_repeat (bool): Whether the correct port is allowed to repeat between trials
            **kwargs:
        """
        super(Free_Water, self).__init__()

        if not stage_block:
            raise Warning('No stage_block Event() was passed, youll need to handle stage progression on your own')
        else:
            self.stage_block = stage_block

        # Fixed parameters
        if isinstance(reward, dict):
            self.reward = reward
        else:
            self.reward         = {'type':'duration',
                                   'value': float(reward)}




        # Variable parameters
        self.target = random.choice(['L', 'C', 'R'])
        self.trial_counter = itertools.count(int(current_trial))
        self.triggers = {}

        # Stage list to iterate
        stage_list = [self.water, self.response]
        self.num_stages = len(stage_list)
        self.stages = itertools.cycle(stage_list)

        # Init hardware
        self.hardware = {}
        self.pin_id = {} # Inverse pin dictionary
        self.init_hardware()

        # Set reward values for solenoids
        # TODO: Super inelegant, implement better with reward manager
        if self.reward['type'] == "volume":
            self.set_reward(vol=self.reward['value'])
        else:
            self.set_reward(duration=self.reward['value'])

        self.allow_repeat = bool(allow_repeat)

    def water(self, *args, **kwargs):
        """
        First stage of task - open a port if it's poked.

        Returns:
            dict: Data dictionary containing::

                'target': ('L', 'C', 'R') - correct response
                'timestamp': isoformatted timestamp
                'trial_num': number of current trial
        """
        self.stage_block.clear()

        # Choose random port
        if self.allow_repeat:
            self.target = random.choice(['L', 'C', 'R'])
        else:
            other_ports = [t for t in ['L', 'C', 'R'] if t is not self.target]
            self.target = random.choice(other_ports)

        # Set triggers and set the target LED to green.
        self.triggers[self.target] = self.hardware['PORTS'][self.target].open
        self.set_leds({self.target: [0, 255, 0]})

        # Return data
        data = {
            'target': self.target,
            'timestamp': datetime.datetime.now().isoformat(),
            'trial_num' : next(self.trial_counter)
        }
        return data


    def response(self):
        """
        Just have to alert the Terminal that the current trial has ended
        and turn off any lights.
        """
        # we just have to tell the Terminal that this trial has ended

        # mebs also turn the light off rl quick
        self.set_leds()

        return {'TRIAL_END':True}



    def end(self):
        """
        When shutting down, release all hardware objects and turn LEDs off.
        """
        for k, v in self.hardware.items():
            for pin, obj in v.items():
                obj.release()





"""
Setup a digital scale...

Warning:
    Not implemented yet
"""

import argparse
import os
import subprocess
import getpass

if __name__ == "__main__":

    if os.getuid() != 0:
        raise Exception("Need to run as root")

    parser = argparse.ArgumentParser(description="Setup a USB-HID scale for the terminal")
    parser.add_argument('-v', '--vendor', help="Vendor ID for scale, default is 0x1446 (stamps.com scale)")
    parser.add_argument('-p', '--product', help="Product ID for scale, default is 0x6a73 (stamps.com scale")

    args = parser.parse_args()

    # Parse Arguments and assign defaults
    if args.vendor:
        vendor = args.vendor
    else:
        vendor = "0x1446"

    if args.product:
        product = args.product
    else:
        product = "0x6a73"

    # make rule string
    rule_string = 'ACTION=="add", SUBSYSTEMS=="usb", BUS=="usb", ATTR{{idVendor}}=="{}", ATTR{{idProduct}}=="{}", MODE=="0666", GROUP=="scale"'.format(vendor, product)

    # create udev rule
    rule_fn = '/etc/udev/rules.d/10-local.rules'
    if os.path.exists(rule_fn):
        mode = "r+"
    else:
        mode = "w"

    with open(rule_fn, mode=mode) as udev_rule:
        udev_rule.write(rule_string)

    print('wrote rule: \n{} to file: \n{}'.format(rule_string, rule_fn))
    print('adding user to "scale" group')

    username = getpass.getuser()
    subprocess.call(['adduser',username,'scale'])


    print('added to group, restarting udev')
    subprocess.call(['sudo', 'groupadd', 'scale'])
    subprocess.call(['sudo', 'service', 'udev', 'restart'])

    print('scale set up! unplug and replug your scale!')








"""
A very simple task: Blink an LED

Written by @mikewehr in the ``mike`` branch: https://github.com/wehr-lab/autopilot/blob/mike/autopilot/tasks/blink.py

Demonstrates the basic structure of a task with one stage,
described in the comments throughout the task.

See the main tutorial for more detail: https://docs.auto-pi-lot.com/en/latest/guide.task.html#

This page is rendered in the docs here in order to provide links to the mentioned objects/classes/etc., but
this example was intended to be read as source code, as some comments will only be visible there.

.. note::
    Currently, after completion, the task needs to be explicitly imported and added to :data:`.tasks.TASK_LIST` ,
    this will be remedied shortly (see the ``registries`` branch).
"""
import itertools
import tables
import time
from datetime import datetime

from autopilot.hardware import gpio
from autopilot.tasks import Task
from collections import OrderedDict as odict

class Blink(Task):
    """
    Blink an LED.

    Note that we subclass the :class:`.tasks.Task` class (``Blink(Task)``) to provide us with some methods
    useful for all Tasks.

    Args:
        pulse_duration (int, float): Duration the LED should be on, in ms
        pulse_interval (int, float): Duration the LED should be off, in ms

    """
    # Tasks need to have a few class attributes defined to be integrated into the rest of the system
    # See here for more about class vs. instance attributes https://www.toptal.com/python/python-class-attributes-an-overly-thorough-guide

    STAGE_NAMES = ["pulse"] # type: list
    """
    An (optional) list or tuple of names of methods that will be used as stages for the task. 
    
    See :attr:`.Blink.stages` for more information
    """

    PARAMS = odict()
    """
    A dictionary that specifies the parameters that control the operation of the task -- each task presumably has some
    range of options that allow slight variations (eg. different stimuli, etc.) on a shared task structure. This
    dictionary specifies each ``PARAM`` as a human-readable ``tag`` and a ``type`` that is used by the gui to 
    create an appropriate input object. For example::
    
        PARAMS['pulse_duration'] = {'tag': 'LED Pulse Duration (ms)', 'type': 'int'}
        
    When instantiated, these params are passed to the ``__init__`` method.
    
    A :class:`collections.OrderedDict` is used so that parameters can be presented in a predictable way to users.
    """
    PARAMS['pulse_duration'] = {'tag': 'LED Pulse Duration (ms)', 'type': 'int'}
    PARAMS['pulse_interval'] = {'tag': 'LED Pulse Interval (ms)', 'type': 'int'}

    class TrialData(tables.IsDescription):
        """
        This class declares the data that will be returned for each "trial" -- or complete set of executed task
        stages. It is used by the :class:`.subject.Subject` object to make a data table with the
        correct data types. Declare each piece of data using a pytables Column descriptor
        (see https://www.pytables.org/usersguide/libref/declarative_classes.html#col-sub-classes for available
        data types, and the pytables guide: https://www.pytables.org/usersguide/tutorials.html for more information)

        For each trial, we'll return two timestamps, the time we turned the LED on, the time we turned it off,
        and the trial number. Note that we use a 26-character :class:`tables.StringCol` for the timestamps,
        which are given as an isoformatted string like ``'2021-02-16T18:11:35.752110'``
        """
        trial_num = tables.Int32Col()
        timestamp_on = tables.StringCol(26)
        timestamp_off = tables.StringCol(26)


    HARDWARE = {
        'LEDS': {
            'dLED': gpio.Digital_Out
        }
    }
    """
    Declare the hardware that will be used in the task. Each hardware object is specified with a ``group`` and 
    an ``id`` as nested dictionaries. These descriptions require a set of hardware parameters in the autopilot
    ``prefs.json`` (typically generated by :mod:`autopilot.setup.setup_autopilot` ) with a matching ``group`` and
    ``id`` structure. For example, an LED declared like this in the :attr:`.Blink.HARDWARE` attribute::
    
        HARDWARE = {'LEDS': {'dLED': gpio.Digital_Out}}
        
    requires an entry in ``prefs.json`` like this::
    
        "HARDWARE": {"LEDS": {"dLED": {
            "pin": 1,
            "polarity": 1
        }}}
        
    that will be used to instantiate the :class:`.hardware.gpio.Digital_Out` object, which is then available for use
    in the task like::
    
        self.hardware['LEDS']['dLED'].set(1)
    """

    def __init__(self, stage_block=None, pulse_duration=100, pulse_interval=500, *args, **kwargs):
        # first we call the superclass ('Task')'s initialization method. All tasks should accept ``*args``
        # and ``**kwargs`` to pass parameters not explicitly specified by subclass up to the superclass.
        super(Blink, self).__init__(*args, **kwargs)

        # store parameters given on instantiation as instance attributes
        self.pulse_duration = int(pulse_duration)
        self.pulse_interval = int(pulse_interval)
        self.stage_block = stage_block # type: "threading.Event"

        # This allows us to cycle through the task by just repeatedly calling self.stages.next()
        self.stages = itertools.cycle([self.pulse])
        """
        Some generator that returns the stage methods that define the operation of the task.
        
        To run a task, the :class:`.pilot.Pilot` object will call each stage function, which can return some dictionary
        of data (see :meth:`.Blink.pulse` ) and wait until some flag (:attr:`.Blink.stage_block` ) is set to compute the
        next stage. Since in this case we want to call the same method (:meth:`.Blink.pulse` ) over and over again,
        we use an :class:`itertools.cycle` object (if we have more than one stage to call in a cycle, we could provide
        them like ``itertools.cycle([self.stage_method_1, self.stage_method_2])`` . More complex tasks can define a custom
        generator for finer control over stage progression.
        """

        self.trial_counter = itertools.count()
        """
        Some counter to keep track of the trial number
        """


        self.init_hardware()
        """
        Hardware is initialized by the superclass's :meth:`.Task.init_hardware` method, which creates all the
        hardware objects defined in :attr:`.Blink.HARDWARE` according to their parameterization in 
        ``prefs.json`` , and makes them available in the :attr:`.Blink.hardware` dictionary.
        """
        self.logger.debug('Hardware initialized')
        """
        All task subclass objects have an :attr:`.Task.logger` -- a :class:`logging.Logger` that allows
        users to easily debug their tasks and see feedback about their operation. To prevent stdout from
        getting clogged, logging messages are printed and stored according to the ``LOGLEVEL`` pref -- so this
        message would only appear if ``LOGLEVEL == "DEBUG"``
        """

        self.stage_block.set()
        """
        We set the stage block and never clear it so that the :class:`.Pilot` doesn't wait for a trigger 
        to call the next stage -- it just does it as soon as the previous one completes. 
        
        See :meth:`.Pilot.run_task` for more detail on this loop.
        """


    ##################################################################################
    # Stage Functions
    ##################################################################################
    def pulse(self, *args, **kwargs):
        """
        Turn an LED on and off according to :attr:`.Blink.pulse_duration` and :attr:`.Blink.pulse_interval`

        Returns:
            dict: A dictionary containing the trial number and two timestamps.
        """
        # -------------
        # turn light on

        # use :meth:`.hardware.gpio.Digital_Out.set` method to turn the LED on
        self.hardware['LEDS']['dLED'].set(1)
        # store the timestamp
        timestamp_on = datetime.now().isoformat()
        # log status as a debug message
        self.logger.debug('light on')
        # sleep for the pulse_duration
        time.sleep(self.pulse_duration / 1000)

        # ------------
        # turn light off, same as turning it on.

        self.hardware['LEDS']['dLED'].set(0)
        timestamp_off = datetime.now().isoformat()
        self.logger.debug('light off')
        time.sleep(self.pulse_interval / 1000)

        # count and store the number of the current trial
        self.current_trial = next(self.trial_counter)


        data = {
            'trial_num': self.current_trial,
            'timestamp_on': timestamp_on,
            'timestamp_off': timestamp_off
        }
        """
        Create the data dictionary to be returned from the stage. Note that each of the keys in the dictionary
        must correspond to the names of the columns declared in the :attr:`.Blink.TrialData` descriptor. 
        
        At the conclusion of running the task, we will be able to access the data from the run with 
        :meth:`.Subject.get_trial_data`, which will be a :class:`pandas.DataFrame` with a row for each trial, and
        a column for each of the fields here.
        """

        # return the data dictionary from the stage method  and yr done :)
        return data



from autopilot import prefs

try:
    import pyo
    PYO = True
except ImportError:
    PYO = False

def pyo_server(debug=False):
    """
    Returns a booted and started pyo audio server

    Warning:
        Use of pyo is generally discouraged due to dropout issues and
        the general opacity of the module.

    Args:
        debug (bool): If true, setVerbosity of pyo server to 8.
    """
    # Jackd should already be running from the launch script created by setup_pilot, we we just
    pyo_server = pyo.Server(audio='jack', nchnls=int(prefs.get('NCHANNELS')),
                            duplex=0, buffersize=4096, sr=192000, ichnls=0)

    # Deactivate MIDI because we don't use it and it's expensive
    pyo_server.deactivateMidi()

    # We have to set pyo to not automatically try to connect to inputs when there aren't any
    pyo_server.setJackAuto(False, True)

    # debug
    if debug:
        pyo_server.setVerbosity(8)

    # Then boot and start
    pyo_server.boot()
    pyo_server.start()

    return pyo_server

